! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                  1D dynamic load balancing                   !
!                                                              !
!--------------------------------------------------------------!

!! 1D dynamic load balancing. Intented to be used in combination with SFC load balancing.
!!
!! Two-stage algorithm:
!!  - 1st stage: Create well-balanced partitioning by comparing the maximum process load
!!    of $maxtries alternative partitions (good value for $maxtries is $nprocs).
!!  - 2nd stage: Shift the process borders around to reduce migration costs.
!!    With $mi_tolerance > 0 the load balance is allowed to decrease when shifting.
!!
!! The algorithm is completely serial.
!!
!! *Preprocessor Options*:
!!  - PARTITION1D_VERBOSE: verbose mode
!!
!! *TODO/IDEAS*:
!!  - recycle used entries in part (free-list)
!!  - parallelization (different processes follow different paths instead of spawn
!!    - problem: global maxload needed
!!
!! @author Matthias Lieber
!!
module partition1d_mod

  use kinds_mod
  implicit none
  private  
  
  public :: partition1d_rebalance

  !! Type for one partition try
  type partition1d_try
    real(r8k)    :: maxload = 0.0    !! max process load
    real(r8k)    :: sumload = 0.0    !! total load partitioned so far
    real(r8k)    :: curload = 0.0    !! load of current process
    integer(i_k) :: pos = 1          !! position in block list ($weights)
    integer(i_k) :: rank = 0         !! current process
  end type partition1d_try
  
contains

!! 1D dynamic load balancing routine.
!! The vector of block weights is devided in $nprocs nearly equal-weighted, consecutive partitions.
!!
!! - Block weights must be >0
!! - -1 for an entry in $blkmap can be used for unassigned blocks (they have no migration costs)
!! - $lb_tolerance should be between 0 and 0.01
!! - $mi_tolerance should be between 0 and 0.1
!!
subroutine partition1d_rebalance(nblocks, nprocs, maxtries, weights, blkmap, lb_tolerance, mi_tolerance, partition)
  integer(i_k), intent(in)  :: nblocks                !! number of blocks
  integer(i_k), intent(in)  :: nprocs                 !! number of processes
  integer(i_k), intent(in)  :: maxtries               !! max. number of different partitions to try out
  real(r4k),    intent(in)  :: weights(nblocks)       !! block weights, must be >0
  integer(i_k), intent(in)  :: blkmap(nblocks)        !! map of current block owners (-1 for unassigned blocks)
  real(r4k),    intent(in)  :: lb_tolerance           !! tolerance for premature exit of loop (0 means none)
  real(r4k),    intent(in)  :: mi_tolerance           !! tolerance for decreasing balance when optimizing migration
  integer(i_k), intent(out) :: partition(0:nprocs)    !! output partition vector, contains start indices of partitions

  ! for 1st stage
  type(partition1d_try), target :: parts(maxtries)
  type(partition1d_try), pointer :: cur
  integer(i_k) :: ipart, top, i
  integer(i_k) :: work(0:nprocs)
  integer(i_k), parameter :: nconv=10
  integer(i_k) :: iconv
  real(r8k) :: maxload, sumload, avgload, avg_remaining_load, conv_maxload
  logical :: advance_rank
  
  ! for 2nd stage
  real(r8k) :: leftload, rightload
  integer(i_k) :: rank, border
  real(r8k) :: load_tolerance

  !
  ! trivial cases
  !
  
  if(nblocks==1) then
    partition(0)=1
    partition(1:nprocs)=2
    return
  end if
  
  if(nprocs==1) then
    partition(0)=1
    partition(1)=nblocks+1
    return
  end if
  
  partition(:) = 0

  ipart = 1
  top = 1
  sumload = sum(weights)
  avgload = sumload / nprocs
  maxload = sumload
  avg_remaining_load = 0.0
  cur => parts(ipart)
  work(:) = 0
  work(0) = 1
  conv_maxload = maxload
  iconv = 0
  
  
  !
  ! 1st Stage: build a well balanced partitioning without caring for migration costs
  !
  
  do
    
    do i=cur%pos,nblocks
      cur%sumload = cur%sumload + weights(i)
      cur%curload = cur%curload + weights(i)
      work(cur%rank+1) = i + 1
      if( cur%curload > cur%maxload ) then
        ! We are creating a new maxload.
        avg_remaining_load = (sumload-(cur%sumload-weights(i)))/MAX(REAL(nprocs-1-cur%rank,r8k),0.00001)
        ! We could advance the rank to avoid the new maxload, maybe this will be better in the end.
        ! Only advance rank, if:
        !  * current rank is not the last rank
        !  * current process load is above average process load
        !  * average remaining process load is less than current best maxload
        !  * current rank has already >= 1 block
        advance_rank = ( cur%rank<nprocs-1 .and. cur%curload>avgload .and. avg_remaining_load<maxload .and. work(cur%rank) < i )
        ! If this new maxload is not worse then the best final maxload, we will follow
        ! both pathes from now on, thus we will spawn a new calculation branch
        if( top<maxtries .and. advance_rank .and. cur%curload < maxload ) then
          ! Ok, spawn a new branch.
          top = top + 1
#ifdef PARTITION1D_VERBOSE          
          print *,ipart,'is spawning', top, ' at ', i ,'  -  avg remaining load:',avg_remaining_load
#endif
          parts(top)%maxload   = cur%maxload
          parts(top)%sumload   = cur%sumload
          parts(top)%rank      = cur%rank + 1
          parts(top)%curload   = weights(i)
          parts(top)%pos       = i + 1
        end if
        ! If the new maxload is worse/equal than best final maxload found so far,
        ! we will advance the rank (if allowed) or cancel the calculation.
        if( advance_rank .and. cur%curload >= maxload ) then
          ! Continue with advanced rank.
#ifdef PARTITION1D_VERBOSE          
          print *,ipart,' is advancing rank at ', i ,'  -  avg remaining load:',avg_remaining_load
#endif
          work(cur%rank+1) = i     ! last rank
          cur%rank = cur%rank + 1  ! increase rank
          work(cur%rank+1) = i + 1 ! current (new) rank
          cur%curload = weights(i) ! set weight for current rank
        elseif( cur%curload >= maxload ) then
          ! Cancel calculation.
          cur%pos=-1
          cur%maxload = MAX(cur%curload,cur%maxload)
#ifdef PARTITION1D_VERBOSE          
          print *,ipart,' canceled at ', i
#endif
          exit
        end if
        ! Set new maxload.
        cur%maxload = MAX(cur%curload,cur%maxload)
      end if  
      ! guess next rank
      if( sumload * (cur%rank+1) < cur%sumload * nprocs .and. cur%rank < nprocs-1) then
        cur%rank = cur%rank + 1
        cur%curload = 0
      end if
    end do
    
    if( cur%maxload < maxload ) then
      ! found new best maxload
      maxload = cur%maxload
      partition = work
#ifdef PARTITION1D_VERBOSE          
      print *,ipart,'finished: maxload / balance',cur%maxload,sumload/(cur%maxload*nprocs)
      print *,ipart,'partition:',partition
#endif
      ! premature exit of loop because maxload does converge
      if(iconv == nconv) then
        if( maxload * (1.0+lb_tolerance) > conv_maxload) then
#ifdef PARTITION1D_VERBOSE          
          print *,'exit: maxload is converging'
#endif
          exit
        end if
        conv_maxload = maxload
        iconv = 0
      end if
      iconv = iconv + 1
    end if
    
    if(cur%pos>0) cur%pos=nblocks

    ipart = top + 1
    do while(ipart>1)
      ipart = ipart - 1
      cur => parts(ipart)
      ! only leave the loop (and continue calculating the branch),
      ! if the maxload of the branch is better then current minumum
      if(cur%pos<nblocks) then
        if(cur%maxload<maxload .and. cur%pos>0) then
          work(cur%rank)=cur%pos-1 ! decrease upper bound of last (!) rank
          exit
        end if
        cur%pos=-1
      end if
    end do
    
    if(cur%pos<0 .or. cur%pos==nblocks) exit

  end do
  
  ! check if the last ranks got no blocks (partition vector never written)
  if( partition(nprocs) == 0 ) then
    ! if yes, we fill the zeros with the max number in the partition vector
    ! to make it valid
    i = nprocs - 1
    do while( partition(i) == 0 .and. i>0 )
      i = i - 1
    end do
    partition(i+1:nprocs) = partition(i)
    !partition(nprocs) = partition(i)
    !partition(i:nprocs-1) = partition(i)-1
  end if  
  

#ifdef PARTITION1D_VERBOSE          
  print *,'branches calculated: ',top
#endif


  !
  ! 2nd Stage: Shift the processor borders around without hurting load balance in order
  !            to reduce the migration costs
  
  ! the maximal tolerated load for a process
  load_tolerance = maxload + avgload * mi_tolerance

#ifdef PARTITION1D_VERBOSE          
  print *,'migration = ',get_migration(blkmap,partition)
#endif
  
  leftload = SUM( weights(partition(0):partition(1)-1) )

  do rank=1,nprocs-1
    border = partition(rank)
    if(border>=nblocks) exit
    rightload = SUM( weights(border:partition(rank+1)-1) )
    ! try shifting to the right
    do while( leftload+weights(border)<=load_tolerance .and. blkmap(border)==rank-1 .and. border < partition(rank+1))
#ifdef PARTITION1D_VERBOSE          
      print *,'right shift at ',border,' from rank ',rank,' to ', rank-1
#endif
      rightload = rightload - weights(border)
      leftload = leftload + weights(border)
      partition(rank) = partition(rank) + 1
      border = partition(rank)
    end do
    ! try shifting to the left
    do while( rightload+weights(border-1)<=load_tolerance .and. blkmap(border-1)==rank .and. border > partition(rank-1) )
#ifdef PARTITION1D_VERBOSE          
      print *,'left shift at ',border,' from rank ',rank-1,' to ', rank
#endif
      partition(rank) = partition(rank) - 1
      border = partition(rank)
      rightload = rightload + weights(border)
      !leftload = leftload - weights(border)
    end do
    leftload = rightload
#ifdef PARTITION1D_VERBOSE
    if(leftload /= SUM( weights(partition(rank):partition(rank+1)-1)) ) then
      print *,'error at', rank, leftload,  SUM( weights(partition(rank):partition(rank+1)-1))
      stop
    end if
#endif
  end do

#ifdef PARTITION1D_VERBOSE          
  print *,'migration = ',get_migration(blkmap,partition)
#endif


end subroutine partition1d_rebalance


!! Return the migration costs in given partition
integer(i_k) function get_migration(blkmap, partition)
  integer(i_k), intent(in) :: blkmap(:)
  integer(i_k), intent(in) :: partition(:)
  !
  integer(i_k) :: i,rank
  get_migration=0
  do i=1,size(blkmap)
    rank=blkmap(i) + 1   ! +1 only because of shifted array bounds
    if(partition(rank)>i .or. i>=partition(rank+1)) get_migration = get_migration + 1
  end do
end function get_migration


!! Return the maxiumum load in given partition
real(r4k) function get_maxload(weights, partition)
  real(r4k),    intent(in) :: weights(:)
  integer(i_k), intent(in) :: partition(:)
  !
  integer(i_k) :: i,nprocs,rank
  real(r4k) :: w
  !
  nprocs=size(partition)-1
  get_maxload=0
  do rank=1,nprocs
    w = 0
    do i=partition(rank),partition(rank+1)-1
      w = w + weights(i)
    end do
    get_maxload = MAX(w,get_maxload)
  end do
end function get_maxload


end module partition1d_mod
